@@ -22,11 +22,9 @@
   use vars qw($VERSION);
 
   use parent qw(HTTP::Server::Simple::CGI);
-
-  use Crypt::Monkeysphere::Validator;
-
-  use Crypt::X509 0.50;
+  require Crypt::X509;
   use Regexp::Common qw /net/;
+  use Convert::ASN1;
   use MIME::Base64;
   use IO::Socket;
   use IO::File;
@@ -35,17 +33,15 @@
   use File::HomeDir;
   use Config::General;
   use Crypt::Monkeysphere::MSVA::MarginalUI;
-  use Crypt::Monkeysphere::Logger;
-  use Crypt::Monkeysphere::Util qw(untaint);
+  use Crypt::Monkeysphere::MSVA::Logger;
   use Crypt::Monkeysphere::MSVA::Monitor;
-  use Crypt::Monkeysphere::OpenPGP;
 
   use JSON;
   use POSIX qw(strftime);
   # we need the version of GnuPG::Interface that knows about pubkey_data, etc:
-  use GnuPG::Interface 0.43;
+  use GnuPG::Interface 0.42.02;
 
-  $VERSION = '0.9.2';
+  $VERSION = '0.8';
 
   my $gnupg = GnuPG::Interface::->new();
   $gnupg->options->quiet(1);
@@ -63,13 +59,23 @@
                                    },
                  );
 
+  my $default_keyserver = 'hkp://pool.sks-keyservers.net';
   my $default_keyserver_policy = 'unlessvalid';
 
-  my $logger = Crypt::Monkeysphere::Logger->new($ENV{MSVA_LOG_LEVEL});
+  my $logger = Crypt::Monkeysphere::MSVA::Logger::->new($ENV{MSVA_LOG_LEVEL});
   sub logger {
     return $logger;
   }
 
+  my $rsa_decoder = Convert::ASN1::->new();
+  $rsa_decoder->prepare(q<
+
+   SEQUENCE {
+        modulus INTEGER,
+        exponent INTEGER
+   }
+          >);
+
   sub net_server {
     return 'Net::Server::MSVA';
   };
@@ -131,6 +137,81 @@
                      };
   }
 
+  sub opensshpubkey2key {
+    my $data = shift;
+    # FIXME: do we care that the label matches the type of key?
+    my ($label, $prop) = split(/ +/, $data);
+
+    my $out = parse_rfc4716body($prop);
+
+    return $out;
+  }
+
+  sub rfc47162key {
+    my $data = shift;
+
+    my @goodlines;
+    my $continuation = '';
+    my $state = 'outside';
+    foreach my $line (split(/\n/, $data)) {
+      last if ($state eq 'body' && $line eq '---- END SSH2 PUBLIC KEY ----');
+      if ($state eq 'outside' && $line eq '---- BEGIN SSH2 PUBLIC KEY ----') {
+        $state = 'header';
+        next;
+      }
+      if ($state eq 'header') {
+        $line = $continuation.$line;
+        $continuation = '';
+        if ($line =~ /^(.*)\\$/) {
+          $continuation = $1;
+          next;
+        }
+        if (! ($line =~ /:/)) {
+          $state = 'body';
+        }
+      }
+      push(@goodlines, $line) if ($state eq 'body');
+    }
+
+    msvalog('debug', "Found %d lines of RFC4716 body:\n%s\n",
+            scalar(@goodlines),
+            join("\n", @goodlines));
+    my $out = parse_rfc4716body(join('', @goodlines));
+
+    return $out;
+  }
+
+  sub parse_rfc4716body {
+    my $data = shift;
+
+    return undef
+      unless defined($data);
+    $data = decode_base64($data) or return undef;
+
+    msvalog('debug', "key properties: %s\n", unpack('H*', $data));
+    my $out = [ ];
+    while (length($data) > 4) {
+      my $size = unpack('N', substr($data, 0, 4));
+      msvalog('debug', "size: 0x%08x\n", $size);
+      return undef if (length($data) < $size + 4);
+      push(@{$out}, substr($data, 4, $size));
+      $data = substr($data, 4 + $size);
+    }
+
+    if ($out->[0] ne "ssh-rsa") {
+      return {error => 'Not an RSA key'};
+    }
+
+    if (scalar(@{$out}) != 3) {
+      return {error => 'Does not contain the right number of bigints for RSA'};
+    }
+
+    return { exponent => Math::BigInt->from_hex('0x'.unpack('H*', $out->[1])),
+             modulus => Math::BigInt->from_hex('0x'.unpack('H*', $out->[2])),
+           } ;
+  }
+
+
   # return an arrayref of processes which we can detect that have the
   # given socket open (the socket is specified with its inode)
   sub getpidswithsocketinode {
@@ -295,7 +376,7 @@
 
     # This is part of a spawned child process.  We don't want the
     # child process to destroy the update monitor when it terminates.
-    $self->{updatemonitor}->forget() if exists $self->{updatemonitor} && defined $self->{updatemonitor};
+    $self->{updatemonitor}->forget();
     my $clientinfo = get_client_info(select);
     my $clientuid = $clientinfo->{uid};
 
@@ -354,6 +435,69 @@
     }
   }
 
+  sub keycomp {
+    my $rsakey = shift;
+    my $gpgkey = shift;
+
+    if ($gpgkey->algo_num != 1) {
+      msvalog('verbose', "Monkeysphere only does RSA keys.  This key is algorithm #%d\n", $gpgkey->algo_num);
+    } else {
+      if ($rsakey->{exponent}->bcmp($gpgkey->pubkey_data->[1]) == 0 &&
+          $rsakey->{modulus}->bcmp($gpgkey->pubkey_data->[0]) == 0) {
+        return 1;
+      }
+    }
+    return 0;
+  }
+
+  sub pem2der {
+    my $pem = shift;
+    my @lines = split(/\n/, $pem);
+    my @goodlines = ();
+    my $ready = 0;
+    foreach my $line (@lines) {
+      if ($line eq '-----END CERTIFICATE-----') {
+        last;
+      } elsif ($ready) {
+        push @goodlines, $line;
+      } elsif ($line eq '-----BEGIN CERTIFICATE-----') {
+        $ready = 1;
+      }
+    }
+    msvalog('debug', "%d lines of base64:\n%s\n", $#goodlines + 1, join("\n", @goodlines));
+    return decode_base64(join('', @goodlines));
+  }
+
+  sub der2key {
+    my $rawdata = shift;
+
+    my $cert = Crypt::X509::->new(cert => $rawdata);
+
+    my $key = {error => 'I do not know what happened here'};
+
+    if ($cert->error) {
+      $key->{error} = sprintf("Error decoding X.509 certificate: %s", $cert->error);
+    } else {
+      msvalog('verbose', "cert subject: %s\n", $cert->subject_cn());
+      msvalog('verbose', "cert issuer: %s\n", $cert->issuer_cn());
+      msvalog('verbose', "cert pubkey algo: %s\n", $cert->PubKeyAlg());
+      msvalog('verbose', "cert pubkey: %s\n", unpack('H*', $cert->pubkey()));
+
+      if ($cert->PubKeyAlg() ne 'RSA') {
+        $key->{error} = sprintf('public key was algo "%s" (OID %s).  MSVA.pl only supports RSA',
+                                $cert->PubKeyAlg(), $cert->pubkey_algorithm);
+      } else {
+        msvalog('debug', "decoding ASN.1 pubkey\n");
+        $key = $rsa_decoder->decode($cert->pubkey());
+        if (! defined $key) {
+          msvalog('verbose', "failed to decode %s\n", unpack('H*', $cert->pubkey()));
+          $key = {error => 'failed to decode the public key'};
+        }
+      }
+    }
+    return $key;
+  }
+
   sub get_keyserver_policy {
     if (exists $ENV{MSVA_KEYSERVER_POLICY} and $ENV{MSVA_KEYSERVER_POLICY} ne '') {
       if ($ENV{MSVA_KEYSERVER_POLICY} =~ /^(always|never|unlessvalid)$/) {
@@ -376,184 +520,63 @@
 
     # FIXME: some msva.conf or monkeysphere.conf file (system and user?)
 
-    # let the keyserver routines choose.
-    return undef;
-  }
-
-
-##################################################
-## PKC KEY EXTRACTION ############################
-
-  sub pkcextractkey {
-    my $data = shift;
-    my $key;
-
-    if (lc($data->{pkc}->{type}) eq 'x509der') {
-      $key = der2key(join('', map(chr, @{$data->{pkc}->{data}})));
-    } elsif (lc($data->{pkc}->{type}) eq 'x509pem') {
-      $key = der2key(pem2der($data->{pkc}->{data}));
-    } elsif (lc($data->{pkc}->{type}) eq 'opensshpubkey') {
-      $key = opensshpubkey2key($data->{pkc}->{data});
-    } elsif (lc($data->{pkc}->{type}) eq 'rfc4716') {
-      $key = rfc47162key($data->{pkc}->{data});
+    # or else read from the relevant gnupg.conf:
+    my $gpghome;
+    if (exists $ENV{GNUPGHOME} and $ENV{GNUPGHOME} ne '') {
+      $gpghome = untaint($ENV{GNUPGHOME});
     } else {
-      $key->{error} = sprintf("Don't know this public key carrier type: %s", $data->{pkc}->{type});
+      $gpghome = File::Spec->catfile(File::HomeDir->my_home, '.gnupg');
     }
-
-    if (exists $key->{error}) {
-      return $key;
-    }
-
-    # make sure that the returned integers are Math::BigInts:
-    $key->{exponent} = Math::BigInt::->new($key->{exponent}) unless (ref($key->{exponent}));
-    $key->{modulus} = Math::BigInt::->new($key->{modulus}) unless (ref($key->{modulus}));
-    msvalog('debug', "pubkey info:\nmodulus: %s\nexponent: %s\n",
-            $key->{modulus}->as_hex(),
-            $key->{exponent}->as_hex(),
-           );
-
-    if ($key->{modulus}->copy()->blog(2) < 1000) {
-      $key->{error} = sprintf('Public key size is less than 1000 bits (was: %d bits)', $key->{modulus}->copy()->blog(2));
-    }
-
-    return $key;
-  }
-
-  sub der2key {
-    my $rawdata = shift;
-
-    my $cert = Crypt::X509::->new(cert => $rawdata);
-
-    my $key = {error => 'I do not know what happened here'};
-
-    if ($cert->error) {
-      $key->{error} = sprintf("Error decoding X.509 certificate: %s", $cert->error);
-    } else {
-      msvalog('verbose', "cert subject: %s\n", $cert->subject_cn());
-      msvalog('verbose', "cert issuer: %s\n", (defined $cert->issuer_cn() ? $cert->issuer_cn() : '<none>'));
-      msvalog('verbose', "cert pubkey algo: %s\n", $cert->PubKeyAlg());
-      msvalog('verbose', "cert pubkey: %s\n", unpack('H*', $cert->pubkey()));
-
-      if ($cert->PubKeyAlg() ne 'RSA') {
-        $key->{error} = sprintf('public key was algo "%s" (OID %s).  MSVA.pl only supports RSA',
-                                $cert->PubKeyAlg(), $cert->pubkey_algorithm);
-      } else {
-        msvalog('debug', "decoding ASN.1 pubkey\n");
-        $key = $cert->pubkey_components();
-        if (! defined $key) {
-          msvalog('verbose', "failed to decode %s\n", unpack('H*', $cert->pubkey()));
-          $key = {error => 'failed to decode the public key'};
+    my $gpgconf = File::Spec->catfile($gpghome, 'gpg.conf');
+    if (-f $gpgconf) {
+      if (-r $gpgconf) {
+        my %gpgconfig = Config::General::ParseConfig($gpgconf);
+        if ($gpgconfig{keyserver} =~ /^(((hkps?|hkpms|finger|ldap):\/\/)?$RE{net}{domain})$/) {
+          msvalog('debug', "Using keyserver %s from the GnuPG configuration file (%s)\n", $1, $gpgconf);
+          return $1;
         } else {
-          # ensure these are Math::BigInts!
-          $key->{exponent} = Math::BigInt::->new($key->{exponent}) unless (ref($key->{exponent}));
-          $key->{modulus} = Math::BigInt::->new($key->{modulus}) unless (ref($key->{modulus}));
-
-          my $pgpext = $cert->PGPExtension();
-          if (defined $pgpext) {
-            $key->{openpgp4fpr} = Crypt::Monkeysphere::OpenPGP::fingerprint($key, $pgpext);
-            msvalog('verbose', "OpenPGP Fingerprint (derived from X.509 cert): 0x%s\n", uc(unpack("H*", $key->{openpgp4fpr})));
-          }
+          msvalog('error', "Not a valid keyserver (from gpg config %s):\n  %s\n", $gpgconf, $gpgconfig{keyserver});
         }
+      } else {
+        msvalog('error', "The GnuPG configuration file (%s) is not readable\n", $gpgconf);
       }
+    } else {
+      msvalog('info', "Did not find GnuPG configuration file while looking for keyserver '%s'\n", $gpgconf);
     }
-    return $key;
+
+    # the default_keyserver
+    return $default_keyserver;
   }
 
-  sub pem2der {
-    my $pem = shift;
-    my @lines = split(/\r?\n/, $pem);
-    my @goodlines = ();
-    my $ready = 0;
-    foreach my $line (@lines) {
-      if ($line eq '-----END CERTIFICATE-----') {
+  sub fetch_uid_from_keyserver {
+    my $uid = shift;
+
+    my $cmd = IO::Handle::->new();
+    my $out = IO::Handle::->new();
+    my $nul = IO::File::->new("< /dev/null");
+
+    my $ks = get_keyserver();
+    msvalog('debug', "start ks query to %s for UserID: %s\n", $ks, $uid);
+    my $pid = $gnupg->wrap_call
+      ( handles => GnuPG::Handles::->new( command => $cmd, stdout => $out, stderr => $nul ),
+        command_args => [ '='.$uid ],
+        commands => [ '--keyserver',
+                      $ks,
+                      qw( --no-tty --with-colons --search ) ]
+      );
+    while (my $line = $out->getline()) {
+      msvalog('debug', "from ks query: (%d) %s", $cmd->fileno, $line);
+      if ($line =~ /^info:(\d+):(\d+)/ ) {
+        $cmd->print(join(' ', ($1..$2))."\n");
+        msvalog('debug', 'to ks query: '.join(' ', ($1..$2))."\n");
         last;
-      } elsif ($ready) {
-        push @goodlines, $line;
-      } elsif ($line eq '-----BEGIN CERTIFICATE-----') {
-        $ready = 1;
       }
     }
-    msvalog('debug', "%d lines of base64:\n%s\n", $#goodlines + 1, join("\n", @goodlines));
-    return decode_base64(join('', @goodlines));
+    # FIXME: can we do something to avoid hanging forever?
+    waitpid($pid, 0);
+    msvalog('debug', "ks query returns %d\n", POSIX::WEXITSTATUS($?));
   }
 
-  sub opensshpubkey2key {
-    my $data = shift;
-    # FIXME: do we care that the label matches the type of key?
-    my ($label, $prop) = split(/ +/, $data);
-
-    my $out = parse_rfc4716body($prop);
-
-    return $out;
-  }
-
-  sub rfc47162key {
-    my $data = shift;
-
-    my @goodlines;
-    my $continuation = '';
-    my $state = 'outside';
-    foreach my $line (split(/\n/, $data)) {
-      last if ($state eq 'body' && $line eq '---- END SSH2 PUBLIC KEY ----');
-      if ($state eq 'outside' && $line eq '---- BEGIN SSH2 PUBLIC KEY ----') {
-        $state = 'header';
-        next;
-      }
-      if ($state eq 'header') {
-        $line = $continuation.$line;
-        $continuation = '';
-        if ($line =~ /^(.*)\\$/) {
-          $continuation = $1;
-          next;
-        }
-        if (! ($line =~ /:/)) {
-          $state = 'body';
-        }
-      }
-      push(@goodlines, $line) if ($state eq 'body');
-    }
-
-    msvalog('debug', "Found %d lines of RFC4716 body:\n%s\n",
-            scalar(@goodlines),
-            join("\n", @goodlines));
-    my $out = parse_rfc4716body(join('', @goodlines));
-
-    return $out;
-  }
-
-  sub parse_rfc4716body {
-    my $data = shift;
-
-    return undef
-      unless defined($data);
-    $data = decode_base64($data) or return undef;
-
-    msvalog('debug', "key properties: %s\n", unpack('H*', $data));
-    my $out = [ ];
-    while (length($data) > 4) {
-      my $size = unpack('N', substr($data, 0, 4));
-      msvalog('debug', "size: 0x%08x\n", $size);
-      return undef if (length($data) < $size + 4);
-      push(@{$out}, substr($data, 4, $size));
-      $data = substr($data, 4 + $size);
-    }
-
-    if ($out->[0] ne "ssh-rsa") {
-      return {error => 'Not an RSA key'};
-    }
-
-    if (scalar(@{$out}) != 3) {
-      return {error => 'Does not contain the right number of bigints for RSA'};
-    }
-
-    return { exponent => Math::BigInt->from_hex('0x'.unpack('H*', $out->[1])),
-             modulus => Math::BigInt->from_hex('0x'.unpack('H*', $out->[2])),
-           } ;
-  }
-
-## PKC KEY EXTRACTION ############################
-##################################################
-
   sub reviewcert {
     my $data  = shift;
     my $clientinfo  = shift;
@@ -566,14 +589,8 @@
                  message => 'Unknown failure',
                };
 
-    # check that there actually is key data
-    if ($data->{pkc}->{data} eq '') {
-      $ret->{message} = sprintf("Key data empty.");
-      return $status,$ret;
-    }
-
     # check context string
-    if ($data->{context} =~ /^(https|ssh|smtp|ike|postgresql|imaps|imap|submission|e-mail)$/) {
+    if ($data->{context} =~ /^(https|ssh|smtp|ike|postgresql|imaps|imap|submission)$/) {
 	$data->{context} = $1;
     } else {
 	msvalog('error', "invalid context: %s\n", $data->{context});
@@ -599,12 +616,11 @@
     }
 
     my $prefix = $data->{context}.'://';
-    if ($data->{context} eq 'e-mail' ||
-       (defined $data->{peer}->{type} &&
+    if (defined $data->{peer}->{type} &&
         $data->{peer}->{type} eq 'client' &&
         # ike and smtp clients are effectively other servers, so we'll
         # exclude them:
-        $data->{context} !~ /^(ike|smtp)$/)) {
+        $data->{context} !~ /^(ike|smtp)$/) {
       $prefix = '';
       # clients can have any one-line User ID without NULL characters
       # and leading or trailing whitespace
@@ -613,7 +629,7 @@
       } else {
         msvalog('error', "invalid client peer name string: %s\n", $data->{peer}->{name});
         $ret->{message} = sprintf("Invalid client peer name string: %s", $data->{peer}->{name});
-        return $status,$ret;
+        return $status, $ret;
       }
     } elsif ($data->{peer}->{name} =~ /^($RE{net}{domain}(:[[:digit:]]+)?)$/) {
       $data->{peer}->{name} = $1;
@@ -629,95 +645,131 @@
     my $uid = $prefix.$data->{peer}->{name};
     msvalog('verbose', "user ID: %s\n", $uid);
 
-    # extract key or openpgp fingerprint from PKC
-    my $fpr;
+    # check pkc type
     my $key;
-    if (lc($data->{pkc}->{type}) eq 'openpgp4fpr') {
-      if ($data->{pkc}->{data} =~ /^(0x)?([[:xdigit:]]{40})$/) {
-	$data->{pkc}->{data} = uc($2);
-	$fpr = $data->{pkc}->{data};
+    if (lc($data->{pkc}->{type}) eq 'x509der') {
+      $key = der2key(join('', map(chr, @{$data->{pkc}->{data}})));
+    } elsif (lc($data->{pkc}->{type}) eq 'x509pem') {
+      $key = der2key(pem2der($data->{pkc}->{data}));
+    } elsif (lc($data->{pkc}->{type}) eq 'opensshpubkey') {
+      $key = opensshpubkey2key($data->{pkc}->{data});
+    } elsif (lc($data->{pkc}->{type}) eq 'rfc4716') {
+      $key = rfc47162key($data->{pkc}->{data});
+    } else {
+      $ret->{message} = sprintf("Don't know this public key carrier type: %s", $data->{pkc}->{type});
+      return $status,$ret;
+    }
+
+    if (exists $key->{error}) {
+      $ret->{message} = $key->{error};
+      return $status,$ret;
+    }
+
+    # make sure that the returned integers are Math::BigInts:
+    $key->{exponent} = Math::BigInt::->new($key->{exponent}) unless (ref($key->{exponent}));
+    $key->{modulus} = Math::BigInt::->new($key->{modulus}) unless (ref($key->{modulus}));
+    msvalog('debug', "pubkey info:\nmodulus: %s\nexponent: %s\n",
+            $key->{modulus}->as_hex(),
+            $key->{exponent}->as_hex(),
+           );
+
+    if ($key->{modulus}->copy()->blog(2) < 1000) {
+      $ret->{message} = sprintf('Public key size is less than 1000 bits (was: %d bits)', $key->{modulus}->copy()->blog(2));
+    } else {
+      $ret->{message} = sprintf('Failed to validate "%s" through the OpenPGP Web of Trust.', $uid);
+      my $lastloop = 0;
+      my $kspolicy;
+      if (defined $data->{keyserverpolicy} &&
+          $data->{keyserverpolicy} =~ /^(always|never|unlessvalid)$/) {
+        $kspolicy = $1;
+        msvalog("verbose", "using requested keyserver policy: %s\n", $1);
       } else {
-	msvalog('error', "invalid OpenPGP v4 fingerprint: %s\n",$data->{pkc}->{data});
-	$ret->{message} = sprintf("Invalid OpenPGP v4 fingerprint.");
-	return $status,$ret;
+        $kspolicy = get_keyserver_policy();
       }
-    } else {
-      # extract key from PKC
-      $key = pkcextractkey($data);
-      if (exists $key->{error}) {
-	$ret->{message} = $key->{error};
-	return $status,$ret;
+      msvalog('debug', "keyserver policy: %s\n", $kspolicy);
+      # needed because $gnupg spawns child processes
+      $ENV{PATH} = '/usr/local/bin:/usr/bin:/bin';
+      if ($kspolicy eq 'always') {
+        fetch_uid_from_keyserver($uid);
+        $lastloop = 1;
+      } elsif ($kspolicy eq 'never') {
+        $lastloop = 1;
       }
-      $fpr = uc(unpack('H*', $key->{openpgp4fpr}))
-        if (exists $key->{openpgp4fpr});
-    }
-    msvalog('verbose', "OpenPGP v4 fingerprint: %s\n",$fpr)
-      if defined $fpr;
+      my $foundvalid = 0;
 
-    # determine keyserver policy
-    my $kspolicy;
-    if (defined $data->{keyserverpolicy} &&
-	$data->{keyserverpolicy} =~ /^(always|never|unlessvalid)$/) {
-      $kspolicy = $1;
-      msvalog("verbose", "using requested keyserver policy: %s\n", $1);
-    } else {
-      $kspolicy = get_keyserver_policy();
-    }
-    msvalog('debug', "keyserver policy: %s\n", $kspolicy);
-    # needed because $gnupg spawns child processes
-    $ENV{PATH} = '/usr/local/bin:/usr/bin:/bin';
+      # fingerprints of keys that are not fully-valid for this User ID, but match
+      # the key from the queried certificate:
+      my @subvalid_key_fprs;
 
-    $ret->{message} = sprintf('Failed to validate "%s" through the OpenPGP Web of Trust.', $uid);
+      while (1) {
+        foreach my $gpgkey ($gnupg->get_public_keys('='.$uid)) {
+          my $validity = '-';
+          foreach my $tryuid ($gpgkey->user_ids) {
+            if ($tryuid->as_string eq $uid) {
+              $validity = $tryuid->validity;
+            }
+          }
+          # treat primary keys just like subkeys:
+          foreach my $subkey ($gpgkey, @{$gpgkey->subkeys}) {
+            my $primarymatch = keycomp($key, $subkey);
+            if ($primarymatch) {
+              if ($subkey->usage_flags =~ /a/) {
+                msvalog('verbose', "key matches, and 0x%s is authentication-capable\n", $subkey->hex_id);
+                if ($validity =~ /^[fu]$/) {
+                  $foundvalid = 1;
+                  msvalog('verbose', "...and it matches!\n");
+                  $ret->{valid} = JSON::true;
+                  $ret->{message} = sprintf('Successfully validated "%s" through the OpenPGP Web of Trust.', $uid);
+                } else {
+                  push(@subvalid_key_fprs, { fpr => $subkey->fingerprint, val => $validity }) if $lastloop;
+                }
+              } else {
+                msvalog('verbose', "key matches, but 0x%s is not authentication-capable\n", $subkey->hex_id);
+              }
+            }
+          }
+        }
+        if ($lastloop) {
+          last;
+        } else {
+          fetch_uid_from_keyserver($uid) if (!$foundvalid);
+          $lastloop = 1;
+        }
+      }
 
-    my $validator=new Crypt::Monkeysphere::Validator(kspolicy=>$kspolicy,
-						     context=>$data->{context},
-						     keyserver=>get_keyserver(),
-						     gnupg=>$gnupg,
-						     logger=>$logger);
-
-    my $uid_query=$validator->lookup(uid=>$uid,fpr=>$fpr,key=>$key);
-
-    # only show the marginal UI if the UID of the corresponding
-    # key is not fully valid.
-    if (defined($uid_query->{valid_key})) {
-      $ret->{valid} = JSON::true;
-      $ret->{message} = sprintf('Successfully validated "%s" through the OpenPGP Web of Trust.', $uid);
-    } else {
-      my $resp = Crypt::Monkeysphere::MSVA::MarginalUI->ask_the_user($gnupg,
-								     $uid,
-								     $uid_query->{subvalid_keys},
-								     getpidswithsocketinode($clientinfo->{inode}),
-								     $logger);
-      msvalog('info', "response: %s\n", $resp);
-      if ($resp) {
-	$ret->{valid} = JSON::true;
-	$ret->{message} = sprintf('Manually validated "%s" through the OpenPGP Web of Trust.', $uid);
+      # only show the marginal UI if the UID of the corresponding
+      # key is not fully valid.
+      if (!$foundvalid) {
+        my $resp = Crypt::Monkeysphere::MSVA::MarginalUI->ask_the_user($gnupg,
+                                                                       $uid,
+                                                                       \@subvalid_key_fprs,
+                                                                       getpidswithsocketinode($clientinfo->{inode}),
+                                                                       $logger);
+        msvalog('info', "response: %s\n", $resp);
+        if ($resp) {
+          $ret->{valid} = JSON::true;
+          $ret->{message} = sprintf('Manually validated "%s" through the OpenPGP Web of Trust.', $uid);
+        }
       }
     }
-
-    return $status,$ret;
+    return $status, $ret;
   }
 
   sub pre_loop_hook {
     my $self = shift;
     my $server = shift;
 
-    $self->spawn_as_child($server);
+    $self->spawn_master_subproc($server);
   }
 
-  sub pre_accept_hook {
+  sub master_subprocess_died {
     my $self = shift;
     my $server = shift;
+    my $subproc_return = shift;
 
-    $self->parent_changed($server) if (defined $self->{parent_pid} && getppid() != $self->{parent_pid});
-  }
-
-  sub parent_changed {
-    my $self = shift;
-    my $server = shift;
-
-    msvalog('verbose', "parent %d went away; exiting.\n", $self->{parent_pid});
-    $server->set_exit_status(0);
+    my $exitstatus = POSIX::WEXITSTATUS($subproc_return);
+    msvalog('verbose', "Subprocess %d terminated; exiting %d.\n", $self->{child_pid}, $exitstatus);
+    $server->set_exit_status($exitstatus);
     $server->server_close();
   }
 
@@ -750,9 +802,22 @@
         # instead, we'll just avoid trying to kill the next process with this PID:
         $self->{updatemonitor}->forget();
       }
+    } elsif (exists $self->{child_pid} &&
+             ($self->{child_pid} == 0 ||
+              $self->{child_pid} == $pid)) {
+      $self->master_subprocess_died($server, $?);
     }
   }
 
+  # use sparingly!  We want to keep taint mode around for the data we
+  # get over the network.  this is only here because we want to treat
+  # the command line arguments differently for the subprocess.
+  sub untaint {
+    my $x = shift;
+    $x =~ /^(.*)$/ ;
+    return $1;
+  }
+
   sub post_bind_hook {
     my $self = shift;
     my $server = shift;
@@ -760,41 +825,36 @@
     $server->{server}->{leave_children_open_on_hup} = 1;
 
     my $socketcount = @{ $server->{server}->{sock} };
-    # note: we're assuming here that if there are more than one socket
-    # open (e.g. IPv6 and IPv4, or multiple IP addresses of the same
-    # family), they all share the same port number as socket 0.
-    if ( $socketcount < 1 ) {
-      msvalog('error', "%d sockets open; should have been at least 1.\n", $socketcount);
+    if ( $socketcount != 1 ) {
+      msvalog('error', "%d sockets open; should have been 1.\n", $socketcount);
       $server->set_exit_status(10);
       $server->server_close();
     }
-    if (!defined($self->port) || $self->port == 0) {
-      my $port = @{ $server->{server}->{sock} }[0]->sockport();
-      if (! defined($port)) {
-        msvalog('error', "got undefined port.\nRecording as 0.\n", $port);
-        $port = 0;
-      } elsif (($port < 1) || ($port >= 65536)) {
-        msvalog('error', "got nonsense port: %d.\nRecording as 0.\n", $port);
-        $port = 0;
-      } elsif ((exists $ENV{MSVA_PORT}) && (($ENV{MSVA_PORT} + 0) != $port)) {
-        msvalog('error', "Explicitly requested port %d, but got port: %d.", ($ENV{MSVA_PORT}+0), $port);
-        $server->set_exit_status(13);
-        $server->server_close();
-      }
-      $self->port($port);
+    my $port = @{ $server->{server}->{sock} }[0]->sockport();
+    if ((! defined $port) || ($port < 1) || ($port >= 65536)) {
+      msvalog('error', "got nonsense port: %d.\n", $port);
+      $server->set_exit_status(11);
+      $server->server_close();
     }
+    if ((exists $ENV{MSVA_PORT}) && (($ENV{MSVA_PORT} + 0) != $port)) {
+      msvalog('error', "Explicitly requested port %d, but got port: %d.", ($ENV{MSVA_PORT}+0), $port);
+      $server->set_exit_status(13);
+      $server->server_close();
+    }
+    $self->port($port);
+    $self->{updatemonitor} = Crypt::Monkeysphere::MSVA::Monitor::->new($logger);
   }
 
-  sub spawn_as_child {
+  sub spawn_master_subproc {
     my $self = shift;
     my $server = shift;
 
-    if ((exists $ENV{MSVA_PARENT_PID}) && ($ENV{MSVA_PARENT_PID} ne '')) {
+    if ((exists $ENV{MSVA_CHILD_PID}) && ($ENV{MSVA_CHILD_PID} ne '')) {
       # this is most likely a re-exec.
-      msvalog('info', "This appears to be a re-exec, continuing with parent pid %d\n", $ENV{MSVA_PARENT_PID});
-      $self->{parent_pid} = $ENV{MSVA_PARENT_PID} + 0;
-     } elsif ($#ARGV >= 0) {
-      $self->{parent_pid} = 0; # indicate that we are planning to fork.
+      msvalog('info', "This appears to be a re-exec, continuing with child pid %d\n", $ENV{MSVA_CHILD_PID});
+      $self->{child_pid} = $ENV{MSVA_CHILD_PID} + 0;
+    } elsif ($#ARGV >= 0) {
+      $self->{child_pid} = 0; # indicate that we are planning to fork.
       # avoid ignoring SIGCHLD right before we fork.
       $SIG{CHLD} = sub {
         my $val;
@@ -802,26 +862,20 @@
           $self->child_dies($val, $server);
         }
       };
-      my $pid = $$;
       my $fork = fork();
       if (! defined $fork) {
         msvalog('error', "could not fork\n");
       } else {
-        if (! $fork) {
-          msvalog('debug', "daemon has PID %d, parent has PID %d\n", $$, $pid);
-          $self->{parent_pid} = $pid;
-          # ppid is set in Net::Server::Fork's post_configure; we're
-          # past post_configure by here, and we're about to change
-          # process IDs before assuming the role of a forking server,
-          # so we should set it properly:
-          $server->{server}->{ppid} = $$;
-          $ENV{MSVA_PARENT_PID} = $pid;
+        if ($fork) {
+          msvalog('debug', "Child process has PID %d\n", $fork);
+          $self->{child_pid} = $fork;
+          $ENV{MSVA_CHILD_PID} = $fork;
         } else {
           msvalog('verbose', "PID %d executing: \n", $$);
           for my $arg (@ARGV) {
             msvalog('verbose', " %s\n", $arg);
           }
-          # untaint the environment for the parent process
+          # untaint the environment for the subprocess
           # see: https://labs.riseup.net/code/issues/2461
           foreach my $e (keys %ENV) {
             $ENV{$e} = untaint($ENV{$e});
@@ -832,22 +886,16 @@
           }
           # restore default SIGCHLD handling:
           $SIG{CHLD} = 'DEFAULT';
-          $ENV{MONKEYSPHERE_VALIDATION_AGENT_SOCKET} = sprintf('http://127.0.0.1:%d', $self->port);
+          $ENV{MONKEYSPHERE_VALIDATION_AGENT_SOCKET} = sprintf('http://localhost:%d', $self->port);
           exec(@args) or exit 111;
         }
       }
     } else {
-      printf("MONKEYSPHERE_VALIDATION_AGENT_SOCKET=http://127.0.0.1:%d;\nexport MONKEYSPHERE_VALIDATION_AGENT_SOCKET;\n", $self->port);
+      printf("MONKEYSPHERE_VALIDATION_AGENT_SOCKET=http://localhost:%d;\nexport MONKEYSPHERE_VALIDATION_AGENT_SOCKET;\n", $self->port);
       # FIXME: consider daemonizing here to behave more like
       # ssh-agent.  maybe avoid backgrounding by setting
       # MSVA_NO_BACKGROUND.
     };
-    if (exists $ENV{MSVA_MONITOR_CHANGES} &&
-        $ENV{MSVA_MONITOR_CHANGES} eq 'true') {
-      $self->{updatemonitor} = Crypt::Monkeysphere::MSVA::Monitor::->new($logger);
-    } else {
-      msvalog('verbose', "Not monitoring for changes\n");
-    }
   }
 
   sub extracerts {
